# db/session.py

from sqlalchemy.orm import sessionmaker
from therix.db.db_manager import DatabaseManager

# Example PostgreSQL connection string
# Format: postgresql://user:password@hostname/database_name
db_manager = DatabaseManager()
SQLALCHEMY_DATABASE_URL = DatabaseManager._construct_db_url()

engine = DatabaseManager()._engine

SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)



from enum import Enum
import json
from therix.utils.rag import get_inference_model
from langchain.chains.llm import LLMChain
from langchain_core.prompts import PromptTemplate
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.text_splitter import CharacterTextSplitter
from langchain.docstore.document import Document
from langchain.chains import (
    MapReduceDocumentsChain,
    ReduceDocumentsChain,
)
from langchain_core.output_parsers import JsonOutputParser
from langfuse.callback import CallbackHandler

       
class SummarizerTypeMaster(Enum): 
    EXTRACTIVE = 'EXTRACTIVE'
    ABSTRACTIVE = 'ABSTRACTIVE'
        
PROMPT_TEMPLATE = {
    "EXTRACTIVE" : """
        Summarize the provided context below:

        {context}

        ---

        Craft your response with conciseness and accuracy, including only the information provided in the context. 
        Use null values for any missing information.

        Please structure your response in the following JSON format:
        {response_schema_json}
        """,
    "ABSTRACTIVE": "Provide a concise summary for the following text using abstractive summarization:\n\n{context}."
}
        
def summarizer(summarizer_config,inference_model_details,text, trace_details, system_prompt=None):
    
    chain_callbacks = []
    if trace_details is not None:
        langfuse_handler = CallbackHandler(
            secret_key=trace_details["secret_key"],
            public_key=trace_details["public_key"],
            host=trace_details["host"],
            trace_name=trace_details["identifier"],
        )
        chain_callbacks.append(langfuse_handler)
        
    pydantic_prompt = summarizer_config['pydantic_model']
    summarization_type = summarizer_config['summarization_type']
    inference_model_name = inference_model_details.name
    inference_model_config = inference_model_details.config
    llm = get_inference_model(inference_model_name, inference_model_config)

    map_template = PromptTemplate.from_template(
        """Read the following document and
        provide a detailed summary of the document.
        {pages}"""
    )
    reduce_template = """The following is set of summaries:
    {pages}
    Take these and distill it into a final, consolidated summary.
    Verify each and every detail and make sure that the final
    summary is accurate and complete."""
    
    map_chain = LLMChain(llm=llm, prompt=map_template)
    reduce_prompt = PromptTemplate.from_template(reduce_template)

    # Run chain
    reduce_chain = LLMChain(llm=llm, prompt=reduce_prompt)

    # Takes a list of documents, combines them into a single string, and passes this to an LLMChain
    combine_documents_chain = StuffDocumentsChain(
        llm_chain=reduce_chain, document_variable_name="pages"
    )

    # Combines and iteratively reduces the mapped documents
    reduce_documents_chain = ReduceDocumentsChain(
        # This is final chain that is called.
        combine_documents_chain=combine_documents_chain,
        # If documents exceed context for `StuffDocumentsChain`
        collapse_documents_chain=combine_documents_chain,
        # The maximum number of tokens to group documents into.
        token_max=4000,
    )

    # Combining documents by mapping a chain over them, then combining results
    map_reduce_chain = MapReduceDocumentsChain(
        # Map chain
        llm_chain=map_chain,
        # Reduce chain
        reduce_documents_chain=reduce_documents_chain,
        # The variable name in the llm_chain to put the documents in
        document_variable_name="pages",
        # Return the results of the map steps in the output
        return_intermediate_steps=False,
    )

    docs = Document(page_content=text, metadata={"source": "local"})

    text_splitter = CharacterTextSplitter.from_tiktoken_encoder(
        chunk_size=1000, chunk_overlap=0
    )
    split_docs = text_splitter.split_documents([docs])

    # Run the chain
    summary = map_reduce_chain.run(split_docs)
    
    parser = JsonOutputParser()

    if(system_prompt and summarization_type=="EXTRACTIVE"):
        prompt = PromptTemplate(
            template=system_prompt.config.get("system_prompt"),
            input_variables=["context"],
            partial_variables={"response_schema_json": pydantic_prompt},
    )
    else: 
        prompt = PromptTemplate(
            template=PROMPT_TEMPLATE[summarization_type],
            input_variables=["context"],
            partial_variables={"response_schema_json": pydantic_prompt},
        )


    
    if summarization_type == SummarizerTypeMaster.EXTRACTIVE.value:
        chain = prompt | llm | parser
        if(trace_details):
            return json.dumps(chain.invoke({"context": summary},config={"callbacks": [langfuse_handler]}))
        else:
            return json.dumps(chain.invoke({"context": summary}))

    else:
        chain = prompt | llm
        if(trace_details):
            return  chain.invoke({"context": summary}, config={"callbacks": [langfuse_handler]}).content
        else:
            return  chain.invoke({"context": summary}).content


from therix.core.constants import DataSourceMaster
from .pipeline_component import DataSource


class PDFDataSource(DataSource):
    def __init__(self, config):
        super().__init__(DataSourceMaster.PDF, config)


class WebsiteDataSource(DataSource):
    def __init__(self, config):
        super().__init__(DataSourceMaster.WEBSITE, config)


class CSVDataSource(DataSource):
    def __init__(self, config):
        super().__init__(DataSourceMaster.CSV, config)


class DatabaseDataSource(DataSource):
    def __init__(self, config):
        super().__init__(DataSourceMaster.DATABASE, config)


class DOCXDataSource(DataSource):
    def __init__(self, config):
        super().__init__(DataSourceMaster.DOCX, config)


class YoutubeDataSource(DataSource):
    def __init__(self, config):
        super().__init__(DataSourceMaster.YOUTUBE, config)



from langchain.prompts.chat import ChatPromptTemplate
from langchain_community.vectorstores.pgvector import PGVector
from therix.utils.rag import get_vectorstore
from langfuse.callback import CallbackHandler

DEFAULT_PROMPT = """
        You are a skilled professional who understands All kinds of documents. 
        You will be provided with Content and keywords.
        You have Analyze given data.
        {{content}}
        {{keywords}}
        Response should be in valid JSON format.

"""

BASIC_INSTRUCTIONS = """
If a question seems confusing, simply rephrase it using simpler words. Examples can help too.
When unsure of the user's intent, rephrase the question and ask for confirmation.
Only answer questions related to your knowledge base.
Avoid asking the user questions unrelated to their current request.
Strive for factual answers. If unsure, acknowledge it and offer to find more information.
Always be polite and inclusive in your communication.
Be open to improvement based on user interactions and feedback.
"""


class KeywordSearchService:
    def __init__(self,llmProvider,embed_model):
        self.llmProvider=llmProvider
        self.embed_model=embed_model

    def keyword_search(self, keyword_search_dict):
        chain_callbacks = []
         
        if keyword_search_dict.get("trace_details") is not None:
            langfuse_handler = CallbackHandler(
            secret_key=keyword_search_dict.get("trace_details")["secret_key"],
            public_key=keyword_search_dict.get("trace_details")["public_key"],
            host=keyword_search_dict.get("trace_details")["host"],
            trace_name=keyword_search_dict.get("trace_details")["identifier"],
        )

        store = get_vectorstore(self.embed_model , str(keyword_search_dict.get("pipeline_id")))

        if(type(keyword_search_dict.get("config_id")) is list):
            retriever=store.as_retriever(
            search_kwargs={"filter": {'configId': {'$in': keyword_search_dict.get('config_id')}}},
            )
        
        else:
            retriever=store.as_retriever(
            search_kwargs={"filter": {'configId': keyword_search_dict.get('config_id')}},
            )
        
        content = ""  
        for keyword in keyword_search_dict["keywords"]:
            results = retriever.get_relevant_documents(keyword)
    
            for res in results:
                content += res.page_content + "\n\n---\n\n" 

        content = content.rstrip("\n\n---\n\n")
        prompt = keyword_search_dict["prompt"]

        if not prompt:
            prompt = DEFAULT_PROMPT.format(content, keyword_search_dict["keywords"])

        prompt = ChatPromptTemplate.from_messages(
            [
                ("system", prompt),
                ("human",
                 "{BASIC_INSTRUCTIONS} {content} {keywords}")
            ]
        )

        chain = prompt | self.llmProvider | keyword_search_dict["output_parser"]
        if(keyword_search_dict.get("trace_details")):
            response_text = chain.invoke({"content": content, "keywords": keyword_search_dict["keywords"], "format_instructions": keyword_search_dict["output_parser"].get_format_instructions(), "BASIC_INSTRUCTIONS": BASIC_INSTRUCTIONS}, config={"callbacks": [langfuse_handler]})
        else: 
            response_text = chain.invoke({"content": content, "keywords": keyword_search_dict["keywords"], "format_instructions": keyword_search_dict["output_parser"].get_format_instructions(), "BASIC_INSTRUCTIONS": BASIC_INSTRUCTIONS})
        
        
        
        return response_text
    
    


import json, uuid
from operator import itemgetter
from pathlib import Path
from urllib import response
from langchain.docstore.document import Document
from langchain_community.document_loaders import TextLoader, PyPDFLoader
from langchain_community.vectorstores.pgvector import PGVector
from langchain_google_genai import ChatGoogleGenerativeAI, GoogleGenerativeAIEmbeddings
from langchain_openai import OpenAIEmbeddings, ChatOpenAI
from langchain_openai import AzureChatOpenAI
from langchain_groq import ChatGroq
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_community.document_loaders.youtube import YoutubeLoader
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import AzureOpenAIEmbeddings
from therix.core.JSONLoader import JSONLoader
from langchain_community.embeddings.bedrock import BedrockEmbeddings
import boto3
from langchain_community.llms import Bedrock
from therix.core.chat_history.base_chat_history import TherixChatMessageHistory
from therix.core.constants import (
    DataSourceMaster,
    EmbeddingModelMaster,
    InferenceModelMaster,
)
from therix.db.session import SQLALCHEMY_DATABASE_URL
from langfuse.callback import CallbackHandler
from langchain_core.prompts import PromptTemplate
from langchain_community.vectorstores.pgvector import PGVector
from langchain.prompts import PromptTemplate
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnablePassthrough
from langchain_core.output_parsers import StrOutputParser
from langchain_core.messages import get_buffer_string
from langchain_core.prompts import format_document
from langchain_core.runnables import RunnableParallel
from langchain_text_splitters import  RecursiveCharacterTextSplitter
from operator import itemgetter
from langchain_experimental.data_anonymizer import PresidioReversibleAnonymizer


def sanitize_text(text):
    # Remove null characters (0x00) from the text
    sanitized_text = text.replace("\x00", "")
    return sanitized_text


def get_loader(file_type, file_path, config):
    if file_type == DataSourceMaster.TEXT:
        return TextLoader(file_path)
    elif file_type == DataSourceMaster.PDF:
        return PyPDFLoader(file_path)
    elif file_type == DataSourceMaster.YOUTUBE:
        return YoutubeLoader.from_youtube_url(
            file_path,
            add_video_info=True,
            language=[config["language"], "id"],
            translation="en",
        )
        return PyPDFLoader(file_path)
    elif file_type == DataSourceMaster.WEBSITE:
        return JSONLoader(file_path)
    else:
        raise ValueError(f"Unknown data source type: {file_type}")


def get_embedding_model(embedding_model_name, config):
    if (
        embedding_model_name == EmbeddingModelMaster.OPENAI_TEXT_ADA
        or embedding_model_name == EmbeddingModelMaster.OPENAI_TEXT_EMBEDDING_3_SMALL
        or embedding_model_name == EmbeddingModelMaster.OPENAI_TEXT_EMBEDDING_3_LARGE
    ) and ("api_key" in config):
        return OpenAIEmbeddings(
            openai_api_key=config["api_key"], model=embedding_model_name
        )
    elif (
        embedding_model_name == EmbeddingModelMaster.AZURE_TEXT_ADA
        or embedding_model_name == EmbeddingModelMaster.AZURE_TEXT_EMBEDDING_3_LARGE
        or embedding_model_name == EmbeddingModelMaster.AZURE_TEXT_EMBEDDING_3_SMALL
    ) and ("azure_api_key" in config):
        return AzureOpenAIEmbeddings(
            api_key=config["azure_api_key"],
            model=embedding_model_name,
            azure_endpoint=config["azure_endpoint"],
            openai_api_version=config["openai_api_version"],
            azure_deployment=config["azure_deployment"],
        )

    elif (
        embedding_model_name == EmbeddingModelMaster.BEDROCK_TITAN_EMBEDDING
        or embedding_model_name
        == EmbeddingModelMaster.BEDROCK_TITAN_MULTIMODAL_EMBEDDING
    ) and ("bedrock_aws_session_token" in config):
        bedrock_client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=config["bedrock_aws_access_key_id"],
            aws_secret_access_key=config["bedrock_aws_secret_access_key"],
            aws_session_token=config["bedrock_aws_session_token"],
            region_name=config["bedrock_region_name"],
        )

        return BedrockEmbeddings(
            credentials_profile_name="bedrock-admin",
            client=bedrock_client,
            model_id=embedding_model_name,
            region_name=config["bedrock_region_name"],
        )
        
    elif(
        embedding_model_name == EmbeddingModelMaster.GEMINI_EMBEDDING
    ) and ("google_api_key" in config):
        return GoogleGenerativeAIEmbeddings(
            model=EmbeddingModelMaster.GEMINI_EMBEDDING,
            google_api_key=config["google_api_key"],
        )
    else:
        raise ValueError(f"Unknown embedding model: {embedding_model_name}")
    
    

def get_inference_model(inference_model_name, config):
    if (
        inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO_INSTRUCT
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_TURBO_PREVIEW
    ) and ("api_key" in config):
        return ChatOpenAI(openai_api_key=config["api_key"], model=inference_model_name)
    elif (
        inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_3_5_TURBO_INSTRUCT
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4
        or inference_model_name == InferenceModelMaster.OPENAI_GPT_4_TURBO_PREVIEW
    ) and ("azure_api_key" in config):
        return AzureChatOpenAI(
            api_key=config["azure_api_key"],
            model=inference_model_name,
            deployment_name=config["azure_deployment"],
            azure_endpoint=config["azure_endpoint"],
            api_version=config["openai_api_version"],
        )
    elif (
        inference_model_name == InferenceModelMaster.GROQ_LLM_MIXTRAL_8_7_B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_LLAMA3_70B
        or inference_model_name == InferenceModelMaster.GROQ_LLM_GEMMA7B
    ) and ("groq_api_key" in config):
        return ChatGroq(groq_api_key=config["groq_api_key"], model=inference_model_name)

    elif (
        inference_model_name == InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1
        or inference_model_name == InferenceModelMaster.BEDROCK_TEXT_LITE_G1
    ) and ("bedrock_aws_session_token" in config):
        bedrock_client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=config["bedrock_aws_access_key_id"],
            aws_secret_access_key=config["bedrock_aws_secret_access_key"],
            aws_session_token=config["bedrock_aws_session_token"],
            region_name=config["bedrock_region_name"],
        )
        return Bedrock(
            credentials_profile_name="bedrock-admin",
            model_id=inference_model_name,
            client=bedrock_client,
            region_name=config["bedrock_region_name"],
        )
    
    elif(
        inference_model_name == InferenceModelMaster.GOOGLE_GEMINI_PRO
        or inference_model_name == InferenceModelMaster.GOOGLE_GEMINI_1_5_PRO
    ) and ("google_api_key" in config):
        return ChatGoogleGenerativeAI(
            model=inference_model_name,
            google_api_key=config["google_api_key"],
            temperature=0.7, 
            top_p=0.85,
            metadata= {'language': 'en'}
          

        )
    else:
        raise ValueError(f"Unknown inference model: {inference_model_name}")


def create_embeddings(data_sources, embedding_model, collection_name):

    store = get_vectorstore(embedding_model, collection_name)
    
    config_id = []
    for data_source in data_sources:
        config_id.append(uuid.uuid4())
        # data_source = {'name': 'PDF', 'config': {'files': ['path/to/file', 'path/to/file']}}
        for file_path in data_source.config["files"]:
            loader = get_loader(data_source.name, file_path, data_source.config)
            extra_metadata = [
                ("configId", str(config_id)),
            ]
            pages = loader.load_and_split()
            for page_content in pages:
                for key, value in extra_metadata:
                    page_content.metadata[key] = value
            
            # document.text = mask_data(document.text)
            result=RecursiveCharacterTextSplitter(
                 chunk_size=1000,
                 chunk_overlap=200,
                ).split_documents(pages)
            store.add_documents(result)
    # config_id is a list, due to possibility of multiple data_source.
    return {"message" : "embedding created" , "config_id" : config_id}


def get_vectorstore(embedding_model, collection_name):
    embeddings = get_embedding_model(embedding_model.name, embedding_model.config)
    store = PGVector(
        collection_name=collection_name,
        connection_string=SQLALCHEMY_DATABASE_URL,
        embedding_function=embeddings,
    )
    return store


def chat(
    question,
    retriever,
    inference_model,
    embed_model,
    session_id,
    pipeline_id,
    trace_details=None,
    pii_filter_config = None,
    system_prompt=None
):
    # print(trace_details)
    chain_callbacks = []
    if trace_details is not None:
        langfuse_handler = CallbackHandler(
            secret_key=trace_details["secret_key"],
            public_key=trace_details["public_key"],
            host=trace_details["host"],
            trace_name=trace_details["identifier"],
        )
        chain_callbacks.append(langfuse_handler)

    history = TherixChatMessageHistory(
        str(session_id),
        str(pipeline_id),
        SQLALCHEMY_DATABASE_URL,
        table_name="chat_history",
    )
    chat_history = history.messages

    if(system_prompt):
        template = system_prompt.config.get("system_prompt")
    else : 
        template = """
        Answer the question based only on the following context:

{context}

Answer the question based on the above context: {question}
        """
    ANSWER_PROMPT = ChatPromptTemplate.from_template(template)
    _template = """Given the following conversation and a follow up question, rephrase the follow up question to be a standalone question, in its original language.
                    Chat History:
                    {chat_history}
                    Follow Up Input: {question}
                    Standalone question:"""

    CONDENSE_QUESTION_PROMPT = PromptTemplate.from_template(_template)
    DEFAULT_DOCUMENT_PROMPT = PromptTemplate.from_template(template="{page_content}")
    
    model = get_inference_model(inference_model.name, inference_model.config)
    
    if pii_filter_config:
        pii_filter_config = pii_filter_config.config
        entities = pii_filter_config['entities']
        anonymizer = PresidioReversibleAnonymizer(
        analyzed_fields=entities,
        faker_seed=42)

        def _combine_documents(
            docs, document_prompt=DEFAULT_DOCUMENT_PROMPT, document_separator="\n\n"
        ):
            doc_strings = [format_document(doc, document_prompt) for doc in docs]
            return anonymizer.anonymize(document_separator.join(doc_strings))
        
        _inputs = RunnableParallel(
        standalone_question=RunnablePassthrough.assign(
            chat_history=lambda x: anonymizer.anonymize(get_buffer_string(x["chat_history"]))
        )
        | CONDENSE_QUESTION_PROMPT
        | model
        | StrOutputParser(),
        )
    else :
        def _combine_documents(
            docs, document_prompt=DEFAULT_DOCUMENT_PROMPT, document_separator="\n\n"
        ):
            doc_strings = [format_document(doc, document_prompt) for doc in docs]
            return document_separator.join(doc_strings)
            
        _inputs = RunnableParallel(
        standalone_question=RunnablePassthrough.assign(
            chat_history=lambda x: get_buffer_string(x["chat_history"])
        )
        | CONDENSE_QUESTION_PROMPT
        | model
        | StrOutputParser(),
        )
    
    _context = {
        "context": itemgetter("standalone_question") | retriever | _combine_documents,
        "question": lambda x: x["standalone_question"],
    }
        
    conversational_qa_chain = _inputs | _context | ANSWER_PROMPT | model
        
    if(trace_details):
        result = conversational_qa_chain.invoke(
            {"question": question, "chat_history": chat_history},
            config={"callbacks": [langfuse_handler]}
        )
    else:
        result = conversational_qa_chain.invoke(
            {"question": question, "chat_history": chat_history}
        )
    
    if inference_model.name == InferenceModelMaster.BEDROCK_TEXT_EXPRES_V1 or inference_model.name == InferenceModelMaster.BEDROCK_TEXT_LITE_G1 : 
        response = f'{result}'
    else :
        response = json.loads(result.json())["content"]

    if pii_filter_config:
        response = anonymizer.deanonymize(response)

    history.add_message("user",question,pipeline_id,session_id)
    history.add_message("system",response,pipeline_id,session_id)
    return response



from typing import Optional
import uuid
from langchain_core.messages import AIMessage
import psycopg
from requests import session
from sqlalchemy import Engine
from therix.db.session import SQLALCHEMY_DATABASE_URL,engine
import logging
from langchain_postgres import PostgresChatMessageHistory
logger = logging.getLogger(__name__)

class TherixChatMessageHistory(PostgresChatMessageHistory):
   
    def __init__(
        self,
        session_id,
        pipeline_id:str,
        engine,
        table_name: str,
    ): 
         self.sync_connection = psycopg.connect(SQLALCHEMY_DATABASE_URL)
         super().__init__(table_name,session_id,sync_connection=self.sync_connection)
         self.session_id = session_id
         self.table_name = table_name
         self.pipeline_id=pipeline_id
         
    
    def add_message(self,message_role, message,pipeline_id,session_id):
         try:
             id=uuid.uuid4()
             cursor=self.sync_connection.cursor()
             insert_query=f"""INSERT INTO {self.table_name} (id, message_role,  message, pipeline_id, session_id) VALUES (%s ,%s, %s, %s, %s);"""
             cursor.execute(insert_query,(id,message_role, message,pipeline_id,session_id))
             self.sync_connection.commit()
             cursor.close()
             
         except (Exception, psycopg.DatabaseError) as e:
             logging.error(e)
            




from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.engine.url import URL
from alembic.config import Config
from alembic import command
import threading
import logging
import os

from therix.entities.models import Base


# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class DatabaseManager:
    _instance = None
    _engine = None
    _session_factory = None

    def __new__(cls):
        logger.debug("Checking instance...")
        if cls._instance is None:
            with threading.Lock():
                if cls._instance is None:
                    logger.debug("Creating new instance...")
                    cls._instance = super(DatabaseManager, cls).__new__(cls)
                    cls._engine = cls._create_engine()
                    cls._session_factory = sessionmaker(
                        bind=cls._engine, autoflush=False, autocommit=False
                    )
                    cls._setup_database()
        return cls._instance

    @classmethod
    def get_session(cls):
        if cls._session_factory is None:
            raise Exception("DatabaseManager is not initialized properly.")
        return scoped_session(cls._session_factory)

    @classmethod
    def _create_engine(cls):
        db_url = cls._construct_db_url()
        return create_engine(db_url)

    @classmethod
    def _construct_db_url(cls):
        drivername = os.getenv("THERIX_DB_TYPE", "postgresql")
        username = os.getenv("THERIX_DB_USERNAME")
        password = os.getenv("THERIX_DB_PASSWORD")
        host = os.getenv("THERIX_DB_HOST")
        port = os.getenv("THERIX_DB_PORT")
        db_name = os.getenv("THERIX_DB_NAME")
        return URL.create(
            drivername=drivername,
            username=username,
            password=password,
            host=host,
            port=port,
            database=db_name,
        ).render_as_string(hide_password=False)

    @staticmethod
    def _setup_database():
        # Placeholder for database setup logic like migrations

        Base.metadata.create_all(DatabaseManager._engine)
        logger.info("Database tables created.")

        alembic_cfg = DatabaseManager._get_alembic_config()
        DatabaseManager._upgrade_database(alembic_cfg)

    @staticmethod
    def _get_alembic_config():
        # Navigate up two levels from the current file location to find the Alembic directory.
        base_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        alembic_config_path = os.path.join(base_dir, "therix/alembic.ini")
        alembic_cfg = Config(alembic_config_path)
        alembic_cfg.set_main_option(
            "script_location", os.path.join(base_dir, "therix/alembic")
        )
        alembic_cfg.set_main_option(
            "sqlalchemy.url", str(DatabaseManager._construct_db_url())
        )
        return alembic_cfg

    @staticmethod
    def _upgrade_database(alembic_cfg):
        try:
            # command.upgrade(alembic_cfg, "head")
            # logger.info("Database upgraded successfully.")
            pass
        except Exception as e:
            logger.error(f"An error occurred while upgrading the database: {e}")
            raise


# Unit testing with pytest will be created next. This would involve testing these functionalities effectively.


db_manager = DatabaseManager()
# Example usage:
# db_manager = DatabaseManager(
#     "postgresql", "postgres", "password", "localhost", "5432", "coditas_dot_ai"
# )
# db_manager1 = DatabaseManager()
# session = db_manager.create_session()
# # Use the session for ORM operations
# db_manager.test_connection()
# session.close()



from therix.core.data_sources import PDFDataSource
from therix.core.inference_models import BedrockLiteG1, BedrockTextExpressV1, GroqMixtral87bInferenceModel
from therix.core.embedding_models import BedrockTitanEmbedding
from therix.core.pii_filter_config import PIIFilterConfig
from therix.core.pipeline import Pipeline
import sys
from therix.core.trace import Trace



# TODO: Init therix with DB details, and license key


## Usage:
# python main.py ad11128d-d2ec-4f7c-8d87-15c1a5dfe1a9 "how does it help in reasoning?"

# if args has pipeline_id, then load the pipeline
## else create new pipeline
if len(sys.argv) > 1:
    pipeline = Pipeline.from_id(sys.argv[1])
    question = sys.argv[2]
    ans = pipeline.invoke(question)
    print(ans)
else:
    pipeline = Pipeline(name="My New Published Pipeline")
    (pipeline
    .add(PDFDataSource(config={'files': ['../../test-data/Essay-on-Lata-Mangeshkar-final.pdf']}))
    .add(BedrockTitanEmbedding(config={ "bedrock_aws_access_key_id":"",
                                        "bedrock_aws_secret_access_key" : "",
                                        "bedrock_aws_session_token" : "",
                                        "bedrock_region_name" : "us-east-1"
                                            }))
    .add(BedrockLiteG1(config={ "bedrock_aws_access_key_id":"",
                                "bedrock_aws_secret_access_key" : "",
                                "bedrock_aws_session_token" : "",
                                "bedrock_region_name" : "us-east-1"
                                            }))
    .add(PIIFilterConfig(config={
        'entities': ['PERSON','PHONE_NUMBER','EMAIL_ADDRESS']
    }))
    .add(Trace(config={
        'secret_key': 'sk-lf-6fd11f23-6506-428c-9328-6a03cce5574b',
        'public_key': 'pk-lf-d319ad68-34e1-45e5-b25c-c341e8f22462',
        'identifier': 'pii_filter_pipeline'
    }))
    .save())

    pipeline.publish()
    pipeline.preprocess_data()
    print(pipeline.id)
    ans = pipeline.invoke("Whom is the data about? And what are their personal details?")

    print(ans)


"""Initial database setup

Revision ID: f5fb392e6b0a
Revises:
Create Date: 2024-04-22 15:11:16.001681

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "f5fb392e6b0a"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    op.create_table(
        "pipelines",
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("status", sa.String(), nullable=True),
        sa.Column("is_deleted", sa.Boolean(), nullable=True),
        sa.Column("deleted_at", sa.DateTime(), nullable=True),
        sa.Column("custom_data", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
    )
    op.create_table(
        "chat_history",
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("message", sa.String(), nullable=False),
        sa.Column("message_role", sa.String(), nullable=False),
        sa.Column("user_id", sa.String(), nullable=False),
        sa.Column("pipeline_id", sa.UUID(), nullable=True),
        sa.Column("session_id", sa.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["pipeline_id"],
            ["pipelines.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
    )
    op.create_table(
        "pipeline_configurations",
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("pipeline_id", sa.UUID(), nullable=True),
        sa.Column(
            "config_type",
            sa.Enum(
                "INPUT_SOURCE",
                "EMBEDDING_MODEL",
                "INFERENCE_MODEL",
                "OUTPUT_SOURCE",
                "TRACE_DETAILS",
                "PII_FILTER",
                "SUMMARIZER",
                name="configtype",
                create_constraint=True,
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("config", sa.JSON(), nullable=True),
        sa.Column("custom_data", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(
            ["pipeline_id"],
            ["pipelines.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
    )
    if not conn.dialect.has_type(conn, "configtype"):
        op.execute(
            "CREATE TYPE configtype AS ENUM('INPUT_SOURCE', 'EMBEDDING_MODEL', 'INFERENCE_MODEL', 'OUTPUT_SOURCE', 'TRACE_DETAILS', 'PII_FILTER', 'SUMMARIZER')"
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("pipeline_configurations")
    op.drop_table("chat_history")
    op.drop_table("pipelines")
    op.execute("DROP TYPE IF EXISTS configtype")
    # ### end Alembic commands ###



class ModelResponse:

    def __init__(self, answer, session_id):
        self.answer = answer
        self.session_id = session_id


   
    def create_response(self):
        return {"answer": self.answer, "session_id": self.session_id}


from therix.utils.summarizer import summarizer
from ..db.session import SessionLocal
from ..entities.models import ConfigType, Pipeline, PipelineConfiguration
import json
import os
import urllib
from therix.services.web_crawling import crawl_website
from therix.utils.pii_filter import pii_filter
from ..db.session import SessionLocal
from therix.utils.rag import chat, create_embeddings, get_vectorstore, get_loader
from langchain_community.chat_models import AzureChatOpenAI
from langchain_openai import AzureOpenAIEmbeddings
from therix.utils.keyword_search import KeywordSearchService


class PipelineService:
    def __init__(self):
        self.db_session = SessionLocal()

    def create_pipeline_with_configurations(self, pipeline_data, configurations_data):
        new_pipeline = Pipeline(**pipeline_data)
        self.db_session.add(new_pipeline)
        self.db_session.flush()  # Flush to assign an ID to the new_pipeline

        for config_data in configurations_data:
            config_data["pipeline_id"] = new_pipeline.id
            new_config = PipelineConfiguration(**config_data)
            self.db_session.add(new_config)

        self.db_session.commit()
        return new_pipeline

    def publish_pipeline(self, pipeline_data):
        pipeline = (
            self.db_session.query(Pipeline).filter_by(id=pipeline_data.id).first()
        )
        pipeline.status = "PUBLISHED"
        self.db_session.commit()
        return pipeline

    def get_pipeline(self, pipeline_id):
        return self.db_session.query(Pipeline).filter_by(id=pipeline_id).first()

    def get_pipeline_configuration(self , pipeline_id: str):
        return (
            self.db_session.query(PipelineConfiguration)
            .filter_by(pipeline_id=pipeline_id)
            .all()
        )

    def get_pipeline_configuraitons_by_type(self, pipeline_id, config_type):
        return (
            self.db_session.query(PipelineConfiguration)
            .filter_by(pipeline_id=pipeline_id, config_type=config_type)
            .all()
        )

    def preprocess_data(self, pipeline_id):
        data_sources = self.get_pipeline_configuraitons_by_type(
            pipeline_id, "INPUT_SOURCE"
        )
        output_file = None
        if "website" in data_sources[0].config:
            website_url = data_sources[0].config["website"]
            web_content = crawl_website(website_url)
            domain_name = urllib.parse.urlparse(website_url).netloc
            output_file = f"{domain_name}_data.json"
            with open(output_file, "w") as f:
                json.dump(web_content, f, indent=4)
            data_sources[0].config["files"] = [output_file]
        if "website" in data_sources[0].config:
            os.remove(output_file)
        embedding_model = self.get_pipeline_configuraitons_by_type(
            pipeline_id, "EMBEDDING_MODEL"
        )
        return create_embeddings(data_sources, embedding_model[0], str(pipeline_id))

    def invoke_pipeline(self, pipeline_id, question, session_id, trace_details=None, system_prompt=None):
        embedding_model = self.get_pipeline_configuraitons_by_type(
            pipeline_id, "EMBEDDING_MODEL"
        )
        store = get_vectorstore(embedding_model[0], str(pipeline_id))
        retreiver = store.as_retriever()

        inference_model = self.get_pipeline_configuraitons_by_type(
            pipeline_id, "INFERENCE_MODEL"
        )
        
        pii_filter_config = self.get_pipeline_configuraitons_by_type(pipeline_id, ConfigType.PII_FILTER)
        if len(pii_filter_config) != 0:
            pii_filter_config = pii_filter_config[0]
        else :
            pii_filter_config = None
                
        result = chat(
            question,
            retreiver,
            inference_model[0],
            embedding_model,
            session_id,
            pipeline_id,
            trace_details,
            pii_filter_config,
            system_prompt
        )
        
        if pii_filter_config:
            pii_filter_config = pii_filter_config.config
            entities = pii_filter_config['entities']
            return pii_filter(result,entities)
        else:
            return result

    def invoke_summarizer_pipeline(self, pipeline_id, text, trace_details = None, system_prompt=None):
        inference_model = self.get_pipeline_configuraitons_by_type(pipeline_id, ConfigType.INFERENCE_MODEL)
        summarizer_config = self.get_pipeline_configuraitons_by_type(pipeline_id, ConfigType.SUMMARIZER)[0].config
        return summarizer(summarizer_config,inference_model[0],text, trace_details, system_prompt)

    def search_keywords(self, keyword_search_params):
        if(not keyword_search_params.get("pipeline_id") or not keyword_search_params.get("config_id") or not keyword_search_params.get("prompt") or not keyword_search_params.get("keywords") or not keyword_search_params.get("output_parser")):
            return "Request is missing required parameters, please provide all the parameters, i.e. pipeline_id, config_id, prompt, keywords, output_parser"
        inference_model = self.get_pipeline_configuraitons_by_type(keyword_search_params.get("pipeline_id"), ConfigType.INFERENCE_MODEL)[0]
        inference_model_name = inference_model.name
        infer_config = inference_model.config
        embedding_model = self.get_pipeline_configuraitons_by_type(keyword_search_params.get("pipeline_id"), ConfigType.EMBEDDING_MODEL)[0]
        embedding_model_name = embedding_model.name
        pipeline_embeddings_model = self.get_pipeline_configuraitons_by_type(keyword_search_params.get("pipeline_id"), ConfigType.EMBEDDING_MODEL)[0]
        keyword_search_params = {
            "infer_config" : infer_config,
            "inference_model": inference_model_name,
            "embedding_model": embedding_model_name,
            "pipeline_embeddings_model": pipeline_embeddings_model,
            "keywords": keyword_search_params.get("keywords"),
            "config_id": keyword_search_params.get("config_id"),
            "output_parser": keyword_search_params.get("output_parser"),
            "prompt" : keyword_search_params.get("prompt"),
            "pipeline_id" : keyword_search_params.get("pipeline_id"),
            "trace_details" : keyword_search_params.get("trace_details")
        }

        AZURE_OPENAI_API_KEY=infer_config.get("azure_api_key")
        AZURE_OPENAI_ENDPOINT=infer_config.get("azure_endpoint")

        embed_config=embedding_model.config

        llm = AzureChatOpenAI(
            deployment_name=keyword_search_params["infer_config"].get("azure_deployment"),
            azure_endpoint=keyword_search_params["infer_config"].get("azure_endpoint"),
            api_version=keyword_search_params["infer_config"].get("openai_api_version"),
            api_key=keyword_search_params["infer_config"].get("azure_api_key"),
            temperature=keyword_search_params["infer_config"].get("temperature") or 0.7
        )

        keyword_search_dict = {
            "pipeline_id" : keyword_search_params.get("pipeline_id"),
            "config_id" : keyword_search_params.get("config_id"),
            "keywords": keyword_search_params.get("keywords"),
            "output_parser": keyword_search_params.get("output_parser"),
            "prompt": keyword_search_params.get("prompt"),
            "trace_details": keyword_search_params.get("trace_details")
        }
        
        return KeywordSearchService(llm, embedding_model).keyword_search(keyword_search_dict)
    
    def __del__(self):
        self.db_session.close()



from uuid import uuid4
from therix.core.pipeline_component import PipelineComponent
from .constants import DataSourceMaster, EmbeddingModelMaster, PipelineTypeMaster  # Import your constants
from ..services.pipeline_service import PipelineService  # Import your service
from ..entities.models import ConfigType  # Import your ConfigType enum
from ..services.pipeline_service import PipelineService  # Import your service
from langchain.globals import set_llm_cache
from therix.core.response import ModelResponse
from sqlalchemy import  create_engine
from therix.core.cache import TherixCache
from therix.db.session import SQLALCHEMY_DATABASE_URL
engine = create_engine(SQLALCHEMY_DATABASE_URL)

class Pipeline:

    def __init__(self, name, status="IN_DRAFT"):
        self.pipeline_data = {"name": name, "status": status}
        self.components = []
        self.type = None
        self.pipeline_service = PipelineService()

    @classmethod
    def from_id(cls, pipeline_id):
        pipeline = cls.__new__(cls)
        pipeline.__init__(None)
        pipeline.load(pipeline_id)
        return pipeline

    def add(self, component):
        if not isinstance(component, PipelineComponent):
            raise ValueError("component must be an instance of PipelineComponent")
        
        component_type_to_pipeline_type = {
            ConfigType.EMBEDDING_MODEL.value: PipelineTypeMaster.RAG.value,
            ConfigType.SUMMARIZER.value: PipelineTypeMaster.SUMMARIZER.value,
            # Add future component types here
        }

        # Check if the component type is in the dictionary and if the pipeline type is not set
        if component.type.value in component_type_to_pipeline_type and self.type is None:
            self.type = component_type_to_pipeline_type[component.type.value]
        elif component.type.value not in component_type_to_pipeline_type:
            pass
        else:
            raise Exception(f'Cannot add {component.type.value}.')
        
        self.components.append(component)
        return self  # Enable method chaining

    def add_data_source(self, name, config):
        data_source = PipelineComponent(ConfigType.INPUT_SOURCE, name, config)
        return self.add(data_source)

    def add_embedding_model(self, name, config):
        embedding_model = PipelineComponent(ConfigType.EMBEDDING_MODEL, name, config)
        return self.add(embedding_model)

    def add_inference_model(self, name, config):
        inference_model = PipelineComponent(ConfigType.INFERENCE_MODEL, name, config)
        return self.add(inference_model)

    def add_output_source(self, name, config):
        output_source = PipelineComponent(ConfigType.OUTPUT_SOURCE, name, config)
        return self.add(output_source)

    def save(self):
        configurations_data = [
            {
                "config_type": component.type.value,
                "name": component.name,
                "config": component.config,
            }
            for component in self.components
        ]
        self.pipeline_data['type'] = self.type
        # Save the pipeline and its components to the database
        self.pipeline_data = self.pipeline_service.create_pipeline_with_configurations(
            self.pipeline_data, configurations_data
        )
        self.id = self.pipeline_data.id
        self.name = self.pipeline_data.name
        return self.pipeline_data

    def publish(self):
        return self.pipeline_service.publish_pipeline(self.pipeline_data)

    def load(self, pipeline_id):
        self.pipeline_data = self.pipeline_service.get_pipeline(pipeline_id)
        self.id = self.pipeline_data.id
        self.name = self.pipeline_data.name
        return self.pipeline_data

    def preprocess_data(self):
       return self.pipeline_service.preprocess_data(self.pipeline_data.id)

    def invoke(self, question=None, session_id=None , keyword_search_params=None):
            cached_response = None
            therix_cache = TherixCache(engine)
            inference_model = self.pipeline_service.get_pipeline_configuraitons_by_type(self.pipeline_data.id, ConfigType.INFERENCE_MODEL)

            if session_id is None:
                session_id = uuid4()
            add_cache = self.pipeline_service.get_pipeline_configuraitons_by_type(self.pipeline_data.id, ConfigType.CACHE_CONFIG)
            
            if(question == None and keyword_search_params == None):
                return "Please provide the required Parameters to invoke the pipeline"

            pipeline_trace_config = self.pipeline_service.get_pipeline_configuraitons_by_type(
                self.pipeline_data.id, ConfigType.TRACE_DETAILS
            )
            pipeline_system_prompt = self.pipeline_service.get_pipeline_configuraitons_by_type(self.pipeline_data.id, ConfigType.SYSTEM_PROMPT)
            trace_details = pipeline_trace_config[0].config if pipeline_trace_config else None

            pipeline_type = self.pipeline_service.get_pipeline(self.pipeline_data.id).type.value

            if pipeline_type == PipelineTypeMaster.RAG.value:
                if add_cache:
                    cached_response = therix_cache.lookup(question, inference_model[0].name, self.pipeline_data.id)
                if cached_response:
                    return ModelResponse(cached_response.response, session_id).create_response()
                if (keyword_search_params and question == None):
                    keyword_search_params['pipeline_id'] = self.pipeline_data.id
                    keyword_search_params['trace_details'] = trace_details
                    answer = self.pipeline_service.search_keywords(keyword_search_params)
                else:
                    if(pipeline_system_prompt):
                        answer = self.pipeline_service.invoke_pipeline(
                            self.pipeline_data.id, question, session_id, trace_details, pipeline_system_prompt[0]
                        )
                    else:
                        answer = self.pipeline_service.invoke_pipeline(
                            self.pipeline_data.id, question, session_id, trace_details
                        )                        
                if add_cache:
                    therix_cache.update(question, inference_model[0].name, answer, self.pipeline_data.id)
            elif pipeline_type == PipelineTypeMaster.SUMMARIZER.value:
                if(pipeline_system_prompt):
                    answer = self.pipeline_service.invoke_summarizer_pipeline(self.pipeline_data.id, text=question, trace_details=trace_details, system_prompt = pipeline_system_prompt[0])
                else: 
                    answer = self.pipeline_service.invoke_summarizer_pipeline(self.pipeline_data.id, text=question, trace_details=trace_details)

            return ModelResponse(answer, session_id).create_response()
            



import json
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from langchain.docstore.document import Document
from langchain.document_loaders.base import BaseLoader


class JSONLoader(BaseLoader):
    def __init__(
        self,
        file_path: Union[str, Path],
        content_key: Optional[str] = None,
    ):
        self.file_path = Path(file_path).resolve()
        self._content_key = content_key

    def create_documents(self, processed_data):
        documents = []
        for item in processed_data:
            content = ''.join(item)
            document = Document(page_content=content, metadata={})
            documents.append(document)
        return documents

    def process_item(self, item, prefix=""):
        if isinstance(item, dict):
            result = []
            for key, value in item.items():
                new_prefix = f"{prefix}.{key}" if prefix else key
                result.extend(self.process_item(value, new_prefix))
            return result
        elif isinstance(item, list):
            result = []
            for value in item:
                result.extend(self.process_item(value, prefix))
            return result
        else:
            return [f"{prefix}: {item}"]

    def process_json(self, data):
        if isinstance(data, list):
            processed_data = []
            for item in data:
                processed_data.extend(self.process_item(item))
            return processed_data
        elif isinstance(data, dict):
            return self.process_item(data)
        else:
            return []

    def load(self) -> List[Document]:
        """Load and return documents from the JSON file."""

        docs = []
        with open(self.file_path, 'r') as json_file:
            try:
                data = json.load(json_file)
                processed_json = self.process_json(data)
                docs = self.create_documents(processed_json)
            except json.JSONDecodeError:
                print("Error: Invalid JSON format in the file.")
        return docs



"""add-type-column-in-pipeline-table

Revision ID: b8fe483b5a71
Revises: 739c7981bdd8
Create Date: 2024-04-26 16:44:38.800600

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = 'b8fe483b5a71'
down_revision: Union[str, None] = '739c7981bdd8'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # op.drop_table('langchain_pg_collection')
    # op.drop_table('langchain_pg_embedding')
    op.add_column('pipelines', sa.Column('type', sa.Enum('RAG', 'SUMMARIZER', name='pipelinetypemaster'), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipelines', 'type')
    op.create_table('langchain_pg_embedding',
    sa.Column('collection_id', sa.UUID(), autoincrement=False, nullable=True),
    sa.Column('embedding', sa.NullType(), autoincrement=False, nullable=True),
    sa.Column('document', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('cmetadata', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
    sa.Column('custom_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('uuid', sa.UUID(), autoincrement=False, nullable=False),
    sa.ForeignKeyConstraint(['collection_id'], ['langchain_pg_collection.uuid'], name='langchain_pg_embedding_collection_id_fkey', ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('uuid', name='langchain_pg_embedding_pkey')
    )
    op.create_table('langchain_pg_collection',
    sa.Column('name', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('cmetadata', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
    sa.Column('uuid', sa.UUID(), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('uuid', name='langchain_pg_collection_pkey')
    )
    # ### end Alembic commands ###



from therix.core.data_sources import PDFDataSource
from typing import List
from therix.core.embedding_models import AzureOpenAIEmbedding3LargeEmbeddingModel
from therix.core.inference_models import AzureOpenAIGPT4InferenceModel
from therix.core.output_parser import OutputParserWrapper
from therix.core.pipeline import Pipeline
import sys
from therix.core.trace import Trace
from pydantic import BaseModel, Field

if len(sys.argv) > 1:
    pipeline = Pipeline.from_id(sys.argv[1])
    question = sys.argv[2]
    session_id = None

    if len(sys.argv) < 4:
        pass
    else:
        session_id = sys.argv[3]

    ans = pipeline.invoke(question, session_id)
    print(ans)
else:
    pipeline = Pipeline(name="My New Published Pipeline")
    (pipeline
    .add(PDFDataSource(config={'files': ['./test-data/file.pdf' , './test-data/rat.pdf']}))
    .add(AzureOpenAIEmbedding3LargeEmbeddingModel(config={"azure_deployment" : "", "azure_api_key" : "" , "azure_endpoint" : "" , "openai_api_version" :  ""}))
    .add(AzureOpenAIGPT4InferenceModel(config={"azure_api_key" : "" , "openai_api_version" :  "", "azure_endpoint" : "" , "azure_deployment" : "" , "temperature" : ""}))
    .add(Trace(config={
                    'secret_key': 'sk-lf-e62aa7ce-c4c8-4c77-ad7d-9d76dfd96db1',
                    'public_key': 'pk-lf-282ad728-c1d6-4247-b6cd-8022198591a9',
                    'identifier': 'keyword_search_pipeline'
         }))
        .save()
        )

    pipeline.publish()
    embedding_created_response = pipeline.preprocess_data()
    print(pipeline.id)
    # ans = pipeline.invoke("What are some use cases of RAT?")
    class TestDetails(BaseModel):
        name: str = Field(description="Name of the Test")
        description: str = Field(description="Short description of the Test")

    class OutputParserJSON(BaseModel):
        tests: List[TestDetails] = Field(description="Test")


    keyword_search_dict = {
        "config_id" : embedding_created_response.get("config_id"),
        "prompt" : "Analyze the provided report content for all the documents provided and include only the sentences that matches {keywords} that is being provided separately and also add the keywords whose records are not found and add description as not found. Response should be a list of JSON  format.{format_instructions} Example: {{tests: [{{name: 'name of test', description: 'short description of Test'}}]}}",
        "keywords" : ["FINE-TUNED MODELS" , "abc" , "RAT"],
        "output_parser" : OutputParserWrapper.parse_output(pydantic_object=OutputParserJSON)
    }
    
    ans = pipeline.invoke(keyword_search_params=keyword_search_dict)
    # ans = pipeline.invoke(question="What are some usecases of RAT", type="RAG")

    print(ans)



import unittest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from therix.core.cache import TherixCache, FulltextLLMCache
from therix.db.session import SQLALCHEMY_DATABASE_URL


class TestTherixCache(unittest.TestCase):
    def setUp(self):
        # Create an in-memory SQLite database for testing
        self.engine = create_engine(SQLALCHEMY_DATABASE_URL)
        # Create tables
        FulltextLLMCache.metadata.create_all(self.engine)
        # Create a session
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # Create an instance of TherixCache
        self.cache = TherixCache(self.engine)

    def tearDown(self):
        # Clean up - close the session and dispose the engine
        self.session.close()
        self.engine.dispose()

    def test_update_and_lookup(self):
        # Test updating cache and looking up for a prompt
        prompt = "Test prompt"
        llm_string = "Test LLM"
        answer = "Test answer"
        pipeline_id = "6f3d36d3-2c43-4d07-b07a-3f3e454f5e01"
        self.cache.update(prompt, llm_string, answer, pipeline_id)
        result = self.cache.lookup(prompt, llm_string, pipeline_id)
        self.assertEqual(result.response, answer)

    def test_invalid_pipeline_id(self):
        # Test updating cache with invalid pipeline ID
        prompt = "Test prompt"
        llm_string = "Test LLM"
        answer = "Test answer"
        invalid_pipeline_id = "invalid_id"
        with self.assertRaises(ValueError):
            self.cache.update(prompt, llm_string, answer, invalid_pipeline_id)

    def test_clear_cache(self):
        # Test clearing cache
        prompt = "Test prompt"
        llm_string = "Test LLM"
        answer = "Test answer"
        pipeline_id = "6f3d36d3-2c43-4d07-b07a-3f3e454f5e01"
        self.cache.update(prompt, llm_string, answer, pipeline_id)
        self.cache.clear()
        result = self.cache.lookup(prompt, llm_string, pipeline_id)   
        self.assertIsNone(result)


if __name__ == '__main__':
    unittest.main()


"""added_config_type_for_system_prompt

Revision ID: f5c4562bcd86
Revises: 1e7f3dc15e9c
Create Date: 2024-05-13 18:42:47.952204

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = 'f5c4562bcd86'
down_revision: Union[str, None] = '1e7f3dc15e9c'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # op.drop_table('langchain_pg_embedding')
    # op.drop_table('llm_cache_fulltext')
    # op.drop_table('langchain_pg_collection')
        op.execute("ALTER TYPE configtype ADD VALUE 'SYSTEM_PROMPT'")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('langchain_pg_collection',
    sa.Column('name', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('cmetadata', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
    sa.Column('uuid', sa.UUID(), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('uuid', name='langchain_pg_collection_pkey'),
    postgresql_ignore_search_path=False
    )
    op.create_table('llm_cache_fulltext',
    sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
    sa.Column('pipeline_id', sa.UUID(), autoincrement=False, nullable=False),
    sa.Column('question', sa.VARCHAR(), autoincrement=False, nullable=False),
    sa.Column('llm', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('response', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.PrimaryKeyConstraint('id', name='llm_cache_fulltext_pkey')
    )
    op.create_table('langchain_pg_embedding',
    sa.Column('collection_id', sa.UUID(), autoincrement=False, nullable=True),
    sa.Column('embedding', sa.NullType(), autoincrement=False, nullable=True),
    sa.Column('document', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('cmetadata', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
    sa.Column('custom_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('uuid', sa.UUID(), autoincrement=False, nullable=False),
    sa.ForeignKeyConstraint(['collection_id'], ['langchain_pg_collection.uuid'], name='langchain_pg_embedding_collection_id_fkey', ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('uuid', name='langchain_pg_embedding_pkey')
    )
    # ### end Alembic commands ###



from therix.core.data_sources import PDFDataSource
from therix.core.embedding_models import (
    BedrockTitanEmbedding,
    OpenAITextAdaEmbeddingModel,
)
from therix.core.inference_models import (
    OpenAIGPT4TurboPreviewInferenceModel,
)
from therix.core.inference_models import (
    OpenAIGPT4TurboPreviewInferenceModel,GroqLlama370b
)
from therix.core.embedding_models import (
    OpenAITextAdaEmbeddingModel,
)
from therix.core.inference_models import (
    OpenAIGPT4TurboPreviewInferenceModel,
)
from therix.core.pipeline import Pipeline
import sys

from therix.core.system_prompt_config import SystemPromptConfig
from therix.core.trace import Trace

GROQ_API_KEY=''

# TODO: Init therix with DB details, and license key


## Usage:
# python main.py ad11128d-d2ec-4f7c-8d87-15c1a5dfe1a9 "how does it help in reasoning?"

# if args has pipeline_id, then load the pipeline
## else create new pipeline
sys_prompt = """Answer the question based only on the following context and reply with your capabilities if something is out of context.
        Context: 
        {context}

        Question: {question}
        """


if len(sys.argv) > 1:
    pipeline = Pipeline.from_id(sys.argv[1])
    question = sys.argv[2]
    session_id = None

    if len(sys.argv) < 4:
        pass
    else:
        session_id = sys.argv[3]

    ans = pipeline.invoke(question, session_id)
    print(ans)
else:
    pipeline = Pipeline(name="My New Published Pipeline")
    (
        pipeline.add(PDFDataSource(config={"files": ["../../test-data/rat.pdf"]}))
        .add(BedrockTitanEmbedding(config={"bedrock_aws_access_key_id" : "",
                                "bedrock_aws_secret_access_key" : ""        ,
                                "bedrock_aws_session_token" : "",
                                "bedrock_region_name" : "us-east-1"}))
        .add(GroqLlama370b(config={"groq_api_key": "GROQ_API_KEY"}))
        .add(SystemPromptConfig(config={"system_prompt" : sys_prompt}))
        
        .add(
            Trace(
                config={
                    "secret_key": "sk-lf-e62aa7ce-c4c8-4c77-ad7d-9d76dfd96db1",
                    "public_key": "pk-lf-282ad728-c1d6-4247-b6cd-8022198591a9",
                    "identifier": "my own pipeline",
                }
            )
        )
        .save()
    )

    pipeline.publish()
    pipeline.preprocess_data()
    print(pipeline.id)
    ans = pipeline.invoke("What are some usecases of RAT?")

    print(ans)
